----------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------
--
--
-- Sintaxe Concreta de IMP2
--
--
----------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------

type Program = Command

data Command = ATTRIB Identifier Expression
             | DECLARE Declaration Command
             | SEQUENCIAL Command Command
             | SKIP
             | CALL Identifier ExpressionList
             | RW IOCommand
	         | IF Expression ThenElseCommand
             | WHILE Expression DoCommand
             deriving Show

data IOCommand = READ Identifier
               | WRITE Expression 
               deriving Show

data DoCommand = DO Command deriving Show
data ThenElseCommand = THEN Command ElseCommand deriving Show
data ElseCommand = ELSE Command deriving Show

data Expression = VALUE Value
                | UNARY UnaryExpression
                | BINARY BinaryExpression
                | ID Identifier
                deriving Show

data Value = INT Int
           | STRING String
           | BOOL Bool
           --| FUNCTION (ListaDecPar, Command)
	   --| ERROR
 	   deriving Show

data UnaryExpression = NOT Expression
                     | LENGTH Expression
                     | MINUS Expression  
                     deriving Show

data BinaryExpression = ADD Expression Expression
                      | SUB Expression Expression
                      | AND Expression Expression
                      | OR Expression Expression
                      | EQUALSTRING Expression Expression
		      | EQUALINT Expression Expression
                      | CONCAT Expression Expression  
                      deriving Show


type Identifier = String

data Declaration = VAR Identifier Expression
                 | PROCEDURE Identifier ParameterList Command
                 | MULTDECL Declaration Declaration
                 deriving Show 


data ExpressionList = SINGLEEXPRESSION Expression
                    | MULTIPLEEXPRESSION Expression ExpressionList
                    deriving Show


data ParameterList = SINGLEPARAMETER Type Identifier                 
                   | MULTIPLEPARAMETER Type Identifier ParameterList
                   deriving Show

data Type = StringType
          | IntType
          | BoolType
          deriving Show

----------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------
--
--
-- Tipos para "simulao do hardware".
--
--
----------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------


type Data = (Identifier, Value)      -- estrutura criada para ligar identificadores a seus repesctivos valores
type Memory = [Data]                 -- estrutura criada para armazenar todos os dados
type Out = String                    -- estrutura criada para armazenar o valor de Out do programa
type BufferReader = Value            -- estrutura de dados que armazenarah

----------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------
--
--
-- Funes para manipulao da memria.
--
--
----------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------

addDeclaration :: Data -> Memory -> Memory
addDeclaration dat mem 
                      | (exists dat mem) == True = mem
                      | otherwise = (dat : mem)

removeDeclaration :: Identifier -> Memory -> Memory
removeDeclaration id [] = []
removeDeclaration (id) ( (i, val) : mem)
                                        | (id == i) = mem
                                        | otherwise = ((i, val) : removeDeclaration (id) mem)

exists :: Data -> Memory -> Bool
exists _ [] = False
exists (i1, val1) ( (i2, val2) : mem) 
                     	             | (i1 == i2) = True
                                     | otherwise = (exists (i1, val1) mem)

getValue :: Identifier -> Memory -> Value
getValue (id) ((i, v) : mem) 
                            | (id == i) = v
                            | otherwise = (getValue (id) mem)

alterValue :: Identifier -> Value -> Memory -> Memory
alterValue _ _ [] = []
alterValue (i) val ((i1, val1) : mem) 
                                   | (i == i1) = ((i1,val) : mem)
                                   | otherwise = ((i1,val1) : (alterValue (i) val mem))

----------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------
--
--
-- Funes de avaliao.
--
--
----------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------
evaluateDeclaration :: Declaration -> Memory -> Memory
evaluateDeclaration (VAR id exp) mem = addDeclaration (id , evaluateExpression exp mem) mem 
evaluateDeclaration (MULTDECL dec1 dec2) mem = evaluateDeclaration dec1 (evaluateDeclaration dec2 mem)

evaluateExpression :: Expression -> Memory -> Value
-- IDENTIFIER
evaluateExpression (ID id) mem = getValue id mem
-- VALUE
evaluateExpression (VALUE (INT i)) mem = (INT i)
evaluateExpression (VALUE (STRING s)) mem = (STRING s)
evaluateExpression (VALUE (BOOL b)) mem = (BOOL b)
-- UNARYEXPRESSION
evaluateExpression (UNARY (MINUS exp)) mem = INT (- (getInt (evaluateExpression exp mem) ) )
evaluateExpression (UNARY (NOT exp)) mem = BOOL (not (getBool (evaluateExpression exp mem) ) )
evaluateExpression (UNARY (LENGTH exp)) mem = INT (length (getString (evaluateExpression exp mem) ) )
-- BINARYEXPRESSION
evaluateExpression (BINARY (ADD exp1 exp2)) mem = INT ((getInt (evaluateExpression exp1 mem)) + (getInt (evaluateExpression exp2 mem)))
evaluateExpression (BINARY (SUB exp1 exp2)) mem = INT ((getInt (evaluateExpression exp1 mem)) - (getInt (evaluateExpression exp2 mem)))
evaluateExpression (BINARY (AND exp1 exp2)) mem = BOOL ((getBool (evaluateExpression exp1 mem)) && (getBool (evaluateExpression exp2 mem)))
evaluateExpression (BINARY (OR exp1 exp2)) mem = BOOL ((getBool (evaluateExpression exp1 mem)) || (getBool (evaluateExpression exp2 mem)))
evaluateExpression (BINARY (CONCAT exp1 exp2)) mem = STRING ((getString (evaluateExpression exp1 mem)) ++ (getString (evaluateExpression exp2 mem)))
evaluateExpression (BINARY (EQUALSTRING exp1 exp2)) mem = BOOL ((getString (evaluateExpression exp1 mem)) == (getString (evaluateExpression exp2 mem)))
evaluateExpression (BINARY (EQUALINT exp1 exp2)) mem = BOOL ((getInt (evaluateExpression exp1 mem)) == (getInt (evaluateExpression exp2 mem)))


evaluateCommand :: Command -> Memory -> Out -> BufferReader -> (Memory , Out, BufferReader)
--ATTRIBUTION
evaluateCommand (ATTRIB id exp) mem out reader = ((alterValue id (evaluateExpression exp mem)  mem) , out, reader)                   
evaluateCommand (DECLARE dec com) mem out reader = evaluateCommand com (evaluateDeclaration dec mem) out reader                            
evaluateCommand (WHILE exp (DO com)) mem out reader 
                                                   | getBool (evaluateExpression exp mem) == False = (mem,out,reader)            
                                                   | otherwise = let
                                                                    (mem1,out1,reader1) = (evaluateCommand com mem out reader)
                                                                 in
                                                                    (evaluateCommand (WHILE exp (DO com)) mem1 out1 reader1)

evaluateCommand (IF exp (THEN com1 (ELSE com2))) mem out reader 
                                                           | (getBool (evaluateExpression exp mem) == True) = (evaluateCommand com1 mem out reader)  
                                                           | otherwise = (evaluateCommand com2 mem out reader)

evaluateCommand (SEQUENCIAL com1 com2) mem out reader = let                                                    
                                                           (mem1,out1,reader1) = (evaluateCommand com1 mem out reader)
                                                        in
                                                           evaluateCommand com2 mem1 out1 reader1
evaluateCommand (SKIP) mem out reader = (mem,out,reader)

evaluateCommand (RW (READ id)) mem out reader = ((alterValue id (getBufferValue reader) mem),out,reader)

evaluateCommand (RW (WRITE exp)) mem out reader = (mem , concatOut (evaluateExpression exp mem) out, reader)


----------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------
--
--
-- Funes auxiliares para leitura, escrita e "desencapsulamento".
--
--
----------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------

getInt :: Value -> Int
getInt (INT i) = i

getString :: Value -> String
getString (STRING s) = s

getBool :: Value -> Bool
getBool (BOOL b) = b

getType :: Type -> String
getType (StringType) = "String"
getType (IntType) = "Int"
getType (BoolType) = "Bool"

getBufferValue :: BufferReader -> Value
getBufferValue reader = reader

concatOut :: Value -> String -> Out
concatOut v s 
    | s == "" = toString v
    |otherwise = (s ++ " , " ++ (toString v))

toString :: Value -> String
toString (INT i) = show i
toString (BOOL b) = show b
toString (STRING s) = show s

getOut :: Out -> String
getOut out = out

run :: Command -> String
run com = let (mem,out,reader) = (evaluateCommand com [] "" (INT 0))
          in getOut out

----------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------
--
--
-- Teste.
--
--
----------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------

prog1 :: Command
com0 :: Command
com1 :: Command
--com2 :: Command
decX :: Declaration
decY :: Declaration

decX = (VAR ("x") (VALUE (INT 10)))
decY = (VAR ("y") (VALUE (INT 10)))
com0 = RW (WRITE (VALUE (STRING "Hello, World!")))
com1 = RW (READ "x")
com2 = RW (WRITE (ID "x"))

--prog1 = SEQUENCIAL (DECLARE decX com) (DECLARE decX com)
--prog1 = SEQUENCIAL com1 com2                
prog1 = SEQUENCIAL (DECLARE decX com1) com0
prog2 :: Command
--prog2 = SEQUENCIAL (DECLARE decX com1) (DECLARE decY com1)
prog2 = SKIP








